import logging
from collections import defaultdict, namedtuple
from math import isclose
from pathlib import Path
from typing import DefaultDict, Iterator, List, Set

import SimpleITK
import numpy as np
import pydicom

from panimg.exceptions import UnconsumedFilesException
from panimg.models import SimpleITKImage

logger = logging.getLogger(__name__)

NUMPY_IMAGE_TYPES = {
    "character": SimpleITK.sitkUInt8,
    "uint8": SimpleITK.sitkUInt8,
    "uint16": SimpleITK.sitkUInt16,
    "uint32": SimpleITK.sitkUInt32,
    "uint64": SimpleITK.sitkUInt64,
    "int8": SimpleITK.sitkInt8,
    "int16": SimpleITK.sitkInt16,
    "int32": SimpleITK.sitkInt32,
    "int64": SimpleITK.sitkInt64,
    "float32": SimpleITK.sitkFloat32,
    "float64": SimpleITK.sitkFloat64,
}

OPTIONAL_METADATA_FIELDS = (
    # These fields will be included in the output mha file
    "Laterality",
    "PatientID",
    "PatientName",
    "PatientBirthDate",
    "PatientAge",
    "PatientSex",
    "StudyDate",
    "StudyInstanceUID",
    "SeriesInstanceUID",
    "StudyDescription",
    "SliceThickness",
    "WindowCenter",
    "WindowWidth",
)


def pixel_data_reached(tag, vr, length):
    return pydicom.datadict.keyword_for_tag(tag) == "PixelData"


def _get_headers_by_study(files, file_errors):
    """
    Gets all headers from dicom files found in path.

    Parameters
    ----------
    path
        Path to a directory that contains all images that were uploaded during
        an upload session.

    Returns
    -------
    A dictionary of sorted headers for all dicom image files found within path,
    grouped by study id.
    """
    studies = {}
    indices = {}

    for file in files:
        if not file.is_file():
            continue
        with file.open("rb") as f:
            try:
                ds = pydicom.filereader.read_partial(
                    f, stop_when=pixel_data_reached
                )
                dims = f"{ds.Rows}x{ds.Columns}"
                key = f"{ds.StudyInstanceUID}-{dims}"
                studies[key] = studies.get(key, {})
                indices[ds.StudyInstanceUID] = indices.get(
                    ds.StudyInstanceUID, {}
                )
                index = indices[ds.StudyInstanceUID].get(dims)
                if index is None:
                    index = (
                        max(list(indices[ds.StudyInstanceUID].values()) + [-1])
                        + 1
                    )
                    indices[ds.StudyInstanceUID][dims] = index
                headers = studies[key].get("headers", [])
                headers.append({"file": file, "data": ds})
                studies[key]["index"] = index
                studies[key]["headers"] = headers
            except Exception as e:
                file_errors[file].append(format_error(str(e)))

    for key in studies:
        studies[key]["headers"].sort(
            key=lambda x: int(x["data"].InstanceNumber)
        )
    return studies


def format_error(message: str) -> str:
    return f"Dicom image builder: {message}"


def _validate_dicom_files(
    files: Set[Path], file_errors: DefaultDict[Path, List[str]]
):
    """
    Gets the headers for all dicom files on path and validates them.

    Parameters
    ----------
    path
        Path to a directory that contains all images that were uploaded during
        an upload session.

    Returns
    -------
    A list of `dicom_dataset` named tuples per study, consisting of:
     - Headers for all dicom image files for the study
     - Number of time points
     - Number of slices per time point

    Any study with an inconsistent amount of slices per time point is discarded.
    """
    studies = _get_headers_by_study(files=files, file_errors=file_errors)
    result = []
    dicom_dataset = namedtuple(
        "dicom_dataset", ["headers", "n_time", "n_slices", "index"]
    )
    for key in studies:
        headers = studies[key]["headers"]
        index = studies[key]["index"]
        if not headers:
            continue
        data = headers[-1]["data"]
        n_time = getattr(data, "TemporalPositionIndex", None)
        n_slices = max(len(headers), int(getattr(data, "NumberOfFrames", 0)))
        # Not a 4d dicom file
        if n_time is None:
            result.append(
                dicom_dataset(
                    headers=headers,
                    n_time=n_time,
                    n_slices=n_slices,
                    index=index,
                )
            )
            continue
        if len(headers) % n_time > 0:
            for d in headers:
                file_errors[d["file"]].append(
                    format_error("Number of slices per time point differs")
                )
            continue
        n_slices = n_slices // n_time
        result.append(
            dicom_dataset(
                headers=headers, n_time=n_time, n_slices=n_slices, index=index,
            )
        )
    del studies
    return result


def _extract_direction(dicom_ds, direction):
    try:
        # Try to extract the direction from the file
        sitk_ref = SimpleITK.ReadImage(str(dicom_ds.headers[0]["file"]))
        # The direction per slice is a 3x3 matrix, so we add the time
        # dimension ourselves
        dims = sitk_ref.GetDimension()
        _direction = np.reshape(sitk_ref.GetDirection(), (dims, dims))
        direction[:dims, :dims] = _direction
    except Exception:
        pass
    return direction


def _process_dicom_file(*, dicom_ds):  # noqa: C901
    ref_file = pydicom.dcmread(str(dicom_ds.headers[0]["file"]))
    ref_origin = tuple(
        float(i) for i in getattr(ref_file, "ImagePositionPatient", (0, 0, 0))
    )
    dimensions = 4 if dicom_ds.n_time and dicom_ds.n_time > 1 else 3
    direction = np.eye(dimensions, dtype=float)
    direction = _extract_direction(dicom_ds, direction)
    pixel_dims = (
        dicom_ds.n_slices,
        int(ref_file.Rows),
        int(ref_file.Columns),
    )
    if dimensions == 4:
        pixel_dims = (dicom_ds.n_time,) + pixel_dims

    # Additional Meta data Contenttimes and Exposures
    content_times = []
    exposures = []
    origin = None
    origin_diff = np.array((0, 0, 0), dtype=float)
    n_diffs = 0
    for partial in dicom_ds.headers:
        ds = partial["data"]
        if "ImagePositionPatient" in ds:
            file_origin = np.array(ds.ImagePositionPatient, dtype=float)
            if origin is not None:
                diff = file_origin - origin
                origin_diff = origin_diff + diff
                n_diffs += 1
            origin = file_origin
    if n_diffs == 0:
        z_i = np.nan
    else:
        avg_origin_diff = tuple(origin_diff / n_diffs)
        try:
            z_i = avg_origin_diff[2]
        except IndexError:
            z_i = 1.0

    samples_per_pixel = int(getattr(ref_file, "SamplesPerPixel", 1))
    img = _create_itk_from_dcm(
        content_times=content_times,
        dicom_ds=dicom_ds,
        dimensions=dimensions,
        exposures=exposures,
        pixel_dims=pixel_dims,
        z_i=z_i,
        samples_per_pixel=samples_per_pixel,
    )

    if origin is None:
        origin = (0.0, 0.0, 0.0)
    sitk_origin = ref_origin if z_i >= 0.0 else tuple(origin)
    z_i = np.abs(z_i) if not np.isnan(z_i) else 1.0

    if "PixelSpacing" in ref_file:
        x_i, y_i = (float(x) for x in ref_file.PixelSpacing)
    else:
        x_i = y_i = 1.0

    sitk_spacing = (x_i, y_i, z_i)
    if dimensions == 4:
        sitk_spacing += (1.0,)
        sitk_origin += (0.0,)

    sitk_direction = tuple(direction.flatten())
    img.SetDirection(sitk_direction)
    img.SetSpacing(sitk_spacing)
    img.SetOrigin(sitk_origin)

    if dimensions == 4:
        # Set Additional Meta Data
        img.SetMetaData("ContentTimes", " ".join(content_times))
        img.SetMetaData("Exposures", " ".join(exposures))

    for f in OPTIONAL_METADATA_FIELDS:
        if getattr(ref_file, f, False):
            img.SetMetaData(f, str(getattr(ref_file, f)))

    return SimpleITKImage(
        image=img,
        name=(
            f"{dicom_ds.headers[0]['data'].StudyInstanceUID}-{dicom_ds.index}"
        ),
        consumed_files={d["file"] for d in dicom_ds.headers},
        spacing_valid=True,
    )


def _create_itk_from_dcm(
    *,
    content_times,
    dicom_ds,
    dimensions,
    exposures,
    pixel_dims,
    z_i,
    samples_per_pixel,
):
    apply_slope = any(
        not isclose(float(getattr(h["data"], "RescaleSlope", 1.0)), 1.0)
        for h in dicom_ds.headers
    )
    apply_intercept = any(
        not isclose(float(getattr(h["data"], "RescaleIntercept", 0.0)), 0.0)
        for h in dicom_ds.headers
    )
    apply_scaling = apply_slope or apply_intercept
    is_rgb = samples_per_pixel > 1
    if apply_scaling:
        np_dtype = np.float32
    else:
        np_dtype = np.short
    if samples_per_pixel > 1:
        pixel_dims += (samples_per_pixel,)
    dcm_array = None
    use_pixel_array = False

    for index, partial in enumerate(dicom_ds.headers):
        ds = pydicom.dcmread(str(partial["file"]))

        if apply_scaling:
            pixel_array = float(
                getattr(ds, "RescaleSlope", 1)
            ) * ds.pixel_array + float(getattr(ds, "RescaleIntercept", 0))
        else:
            pixel_array = ds.pixel_array

        if (
            len(pixel_array.shape) == dimensions
            or pixel_array.shape == pixel_dims
        ):
            use_pixel_array = True
            del ds
            break
        if dcm_array is None:
            dcm_array = np.zeros(pixel_dims, dtype=np_dtype)

        z_index = index if z_i >= 0 else len(dicom_ds.headers) - index - 1

        if dimensions == 4:
            dcm_array[
                index // dicom_ds.n_slices, z_index % dicom_ds.n_slices, :, :
            ] = pixel_array
            if index % dicom_ds.n_slices == 0:
                content_times.append(str(ds.ContentTime))
                exposures.append(str(ds.Exposure))
        else:
            dcm_array[z_index % dicom_ds.n_slices, :, :] = pixel_array

        del ds
    if use_pixel_array:
        img = SimpleITK.GetImageFromArray(pixel_array, isVector=is_rgb)
    else:
        img = SimpleITK.GetImageFromArray(dcm_array, isVector=is_rgb)
    return img


def image_builder_dicom(*, files: Set[Path]) -> Iterator[SimpleITKImage]:
    """
    Constructs image objects by inspecting files in a directory.

    Parameters
    ----------
    path
        Path to a directory that contains all images that were uploaded during
        an upload session.

    Returns
    -------
    An `ImageBuilder` object consisting of:
     - a list of filenames for all files consumed by the image builder
     - a list of detected images
     - a list files associated with the detected images
     - path->error message map describing what is wrong with a given file
    """
    file_errors: DefaultDict[Path, List[str]] = defaultdict(list)

    studies = _validate_dicom_files(files=files, file_errors=file_errors)

    for dicom_ds in studies:
        try:
            yield _process_dicom_file(dicom_ds=dicom_ds)
        except Exception as e:
            for d in dicom_ds.headers:
                file_errors[d["file"]].append(format_error(str(e)))

    if file_errors:
        raise UnconsumedFilesException(file_errors=file_errors)



from construct import PaddedString, Struct, Int32un
import numpy as np
from panimg.contrib.oct_converter.image_types import (
    OCTVolumeWithMetaData,
    FundusImageWithMetaData,
)
from pathlib import Path


class FDS:
    """ Class for extracting data from Topcon's .fds file format.

        Notes:
            Mostly based on description of .fds file format here:
            https://bitbucket.org/uocte/uocte/wiki/Topcon%20File%20Format

        Attributes:
            filepath (str): Path to .img file for reading.
            header (obj:Struct): Defines structure of volume's header.
            oct_header (obj:Struct): Defines structure of OCT header.
            fundus_header (obj:Struct): Defines structure of fundus header.
            chunk_dict (dict): Name of data chunks present in the file, and their start locations.
    """

    def __init__(self, filepath):
        self.filepath = Path(filepath)
        if not self.filepath.exists():
            raise FileNotFoundError(self.filepath)

        self.header = Struct(
            "FOCT" / PaddedString(4, "ascii"),
            "FDA" / PaddedString(3, "ascii"),
            "version_info_1" / Int32un,
            "version_info_2" / Int32un,
        )
        self.oct_header = Struct(
            "unknown" / PaddedString(1, "ascii"),
            "width" / Int32un,
            "height" / Int32un,
            "bits_per_pixel" / Int32un,
            "number_slices" / Int32un,
            "unknown" / PaddedString(1, "ascii"),
            "size" / Int32un,
        )
        self.fundus_header = Struct(
            "width" / Int32un,
            "height" / Int32un,
            "bits_per_pixel" / Int32un,
            "number_slices" / Int32un,
            "unknown" / PaddedString(1, "ascii"),
            "size" / Int32un,
        )
        self.chunk_dict = self.get_list_of_file_chunks()

    def get_list_of_file_chunks(self):
        """Find all data chunks present in the file.

        Returns:
            dict
        """
        chunk_dict = {}
        with open(self.filepath, "rb") as f:
            # skip header
            raw = f.read(15)
            header = self.header.parse(raw)

            eof = False
            while not eof:
                chunk_name_size = np.frombuffer(f.read(1), dtype=np.uint8)[0]
                if chunk_name_size == 0:
                    eof = True
                else:
                    chunk_name = f.read(chunk_name_size)
                    chunk_size = np.frombuffer(f.read(4), dtype=np.uint32)[0]
                    chunk_location = f.tell()
                    f.seek(chunk_size, 1)
                    chunk_dict[chunk_name] = [chunk_location, chunk_size]
        print(f"File {self.filepath} contains the following chunks:")
        for key in chunk_dict.keys():
            print(key)
        return chunk_dict

    def read_oct_volume(self):
        """ Reads OCT data.

            Returns:
                obj:OCTVolumeWithMetaData
        """
        if b"@IMG_SCAN_03" not in self.chunk_dict:
            raise ValueError(
                "Could not find OCT header @IMG_SCAN_03 in chunk list"
            )
        with open(self.filepath, "rb") as f:
            chunk_location, chunk_size = self.chunk_dict[b"@IMG_SCAN_03"]
            f.seek(chunk_location)
            raw = f.read(22)
            oct_header = self.oct_header.parse(raw)
            number_pixels = (
                oct_header.width * oct_header.height * oct_header.number_slices
            )
            raw_volume = np.frombuffer(
                f.read(number_pixels * 2), dtype=np.uint16
            )
            volume = np.array(raw_volume)
            volume = volume.reshape(
                oct_header.width,
                oct_header.height,
                oct_header.number_slices,
                order="F",
            )
            volume = np.transpose(volume, [1, 0, 2])
        oct_volume = OCTVolumeWithMetaData(
            [volume[:, :, i] for i in range(volume.shape[2])]
        )
        return oct_volume

    def read_fundus_image(self):
        """ Reads fundus image.

            Returns:
                obj:FundusImageWithMetaData
        """
        if b"@IMG_OBS" not in self.chunk_dict:
            raise ValueError(
                "Could not find fundus header @IMG_OBS in chunk list"
            )
        with open(self.filepath, "rb") as f:
            chunk_location, chunk_size = self.chunk_dict[b"@IMG_OBS"]
            f.seek(chunk_location)
            raw = f.read(21)
            fundus_header = self.fundus_header.parse(raw)
            # number_pixels = fundus_header.width * fundus_header.height * fundus_header.number_slices
            raw_image = np.frombuffer(
                f.read(fundus_header.size), dtype=np.uint8
            )
            # raw_image = [struct.unpack('B', f.read(1)) for pixel in range(fundus_header.size)]
            image = np.array(raw_image)
            image = image.reshape(
                3, fundus_header.width, fundus_header.height, order="F"
            )
            image = np.transpose(image, [2, 1, 0])
            image = image.astype(np.float32)
        fundus_image = FundusImageWithMetaData(image)
        return fundus_image



from pathlib import Path
from typing import Iterator, Set, Union

from typing_extensions import Protocol  # for py37 support

from panimg.models import (
    PanImgFile,
    PostProcessorResult,
    SimpleITKImage,
    TIFFImage,
)


class ImageBuilder(Protocol):
    def __call__(
        self, *, files: Set[Path]
    ) -> Union[Iterator[SimpleITKImage], Iterator[TIFFImage]]:
        ...


class PostProcessor(Protocol):
    def __call__(self, *, image_files: Set[PanImgFile]) -> PostProcessorResult:
        ...



class FundusImageWithMetaData:
    """ Class to hold the fundus image and any related metadata, and enable saving.

    Attributes:
        image (np.array): Fundus image.
        laterality (str): Left or right eye.
        patient_id (str): Patient ID.
        DOB (str): Patient date of birth.
    """

    def __init__(
        self, image, laterality=None, patient_id=None, patient_dob=None
    ):
        self.image = image
        self.laterality = laterality
        self.patient_id = patient_id
        self.DOB = patient_dob



import re
import zlib
from pathlib import Path
from typing import Any, Dict, List, Mapping, Tuple, Union

import SimpleITK
import SimpleITK._SimpleITK as _SimpleITK

METAIO_IMAGE_TYPES = {
    "MET_NONE": None,
    "MET_ASCII_CHAR": None,
    "MET_CHAR": SimpleITK.sitkInt8,
    "MET_UCHAR": SimpleITK.sitkUInt8,
    "MET_SHORT": SimpleITK.sitkInt16,
    "MET_USHORT": SimpleITK.sitkUInt16,
    "MET_INT": SimpleITK.sitkInt32,
    "MET_UINT": SimpleITK.sitkUInt32,
    "MET_LONG": SimpleITK.sitkInt64,
    "MET_ULONG": SimpleITK.sitkUInt64,
    "MET_LONG_LONG": None,
    "MET_ULONG_LONG": None,
    "MET_FLOAT": SimpleITK.sitkFloat32,
    "MET_DOUBLE": SimpleITK.sitkFloat64,
    "MET_STRING": None,
    "MET_CHAR_ARRAY": SimpleITK.sitkVectorInt8,
    "MET_UCHAR_ARRAY": SimpleITK.sitkVectorUInt8,
    "MET_SHORT_ARRAY": SimpleITK.sitkVectorInt16,
    "MET_USHORT_ARRAY": SimpleITK.sitkVectorUInt16,
    "MET_INT_ARRAY": SimpleITK.sitkVectorInt32,
    "MET_UINT_ARRAY": SimpleITK.sitkVectorUInt32,
    "MET_LONG_ARRAY": SimpleITK.sitkVectorInt64,
    "MET_ULONG_ARRAY": SimpleITK.sitkVectorUInt64,
    "MET_LONG_LONG_ARRAY": None,
    "MET_ULONG_LONG_ARRAY": None,
    "MET_FLOAT_ARRAY": SimpleITK.sitkVectorFloat32,
    "MET_DOUBLE_ARRAY": SimpleITK.sitkVectorFloat64,
    "MET_FLOAT_MATRIX": None,
    "MET_OTHER": None,
}

FLOAT_MATCH_REGEXP = re.compile(r"^[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?$")
FLOAT_LIST_MATCH_REGEXP = re.compile(
    r"^([-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"
    r"(\s[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)*$"
)
CONTENT_TIMES_LIST_MATCH_REGEXP = re.compile(
    r"^((2[0-3]|[0-1]\d)[0-5]\d[0-5]\d(\.\d\d\d)?)"
    r"(\s(2[0-3]|[0-1]\d)[0-5]\d[0-5]\d(\.\d\d\d)?)*$"
)

LENGTH_LIMIT_MATCH_REGEXP = re.compile(r"^.{0,128}$")

STUDYDATE_MATCH_REGEXP = re.compile(r"^\d{4}\d{1,2}\d{1,2}$")

ADDITIONAL_HEADERS = {
    "Laterality": LENGTH_LIMIT_MATCH_REGEXP,
    "PatientID": LENGTH_LIMIT_MATCH_REGEXP,
    "PatientName": LENGTH_LIMIT_MATCH_REGEXP,
    "PatientBirthDate": LENGTH_LIMIT_MATCH_REGEXP,
    "PatientAge": LENGTH_LIMIT_MATCH_REGEXP,
    "PatientSex": LENGTH_LIMIT_MATCH_REGEXP,
    "StudyDate": STUDYDATE_MATCH_REGEXP,
    "StudyInstanceUID": LENGTH_LIMIT_MATCH_REGEXP,
    "SeriesInstanceUID": LENGTH_LIMIT_MATCH_REGEXP,
    "StudyDescription": LENGTH_LIMIT_MATCH_REGEXP,
    "SliceThickness": FLOAT_MATCH_REGEXP,
    "Exposures": FLOAT_LIST_MATCH_REGEXP,
    "ContentTimes": CONTENT_TIMES_LIST_MATCH_REGEXP,
    "WindowCenter": FLOAT_MATCH_REGEXP,
    "WindowWidth": FLOAT_MATCH_REGEXP,
    "t0": FLOAT_MATCH_REGEXP,
    "t1": FLOAT_MATCH_REGEXP,
}

HEADERS_MATCHING_NUM_TIMEPOINTS: List[str] = ["Exposures", "ContentTimes"]

EXPECTED_HEADERS: List[str] = [
    "ObjectType",
    "NDims",
    "BinaryData",
    "BinaryDataByteOrderMSB",
    "CompressedData",
    "CompressedDataSize",
    "TransformMatrix",
    "Offset",
    "CenterOfRotation",
    "AnatomicalOrientation",
    "ElementSpacing",
    "ElementNumberOfChannels",
    "DimSize",
    "ElementType",
    "ElementDataFile",
]


def parse_mh_header(file: Path) -> Dict[str, str]:
    """
    Attempts to parse the headers of an mhd file.

    This function must be secure to safeguard against any untrusted uploaded
    file.

    Parameters
    ----------
    filename

    Returns
    -------
        The extracted header from the mhd file as key value pairs.

    Raises
    ------
    ValueError
        Raised when the file contains problems making it impossible to
        read.
    """

    # attempt to limit number of read headers to prevent overflow attacks
    read_line_limit = 10000

    result: Dict[str, str] = {}
    with file.open("rb") as f:
        lines = True
        while lines:
            read_line_limit -= 1
            if read_line_limit < 0:
                raise ValueError("Files contains too many header lines")

            bin_line = f.readline(10000)

            if not bin_line:
                lines = False
                continue

            if len(bin_line) >= 10000:
                raise ValueError("Line length is too long")

            try:
                line = bin_line.decode("utf-8")
            except UnicodeDecodeError:
                raise ValueError("Header contains invalid UTF-8")
            else:
                result.update(extract_key_value_pairs(line))

            if "ElementDataFile" in result:
                break  # last parsed header...

    return result


def extract_key_value_pairs(line: str) -> Dict[str, str]:
    line = line.rstrip("\n\r")
    if line.strip() and "=" in line:
        key, value = line.split("=", 1)
        return {key.strip(): value.strip()}
    else:
        return {}


def extract_header_listing(
    property: str, headers: Dict[str, str], dtype: type = float
) -> List[Any]:
    return [dtype(e) for e in headers[property].strip().split(" ")]


def load_sitk_image_with_nd_support(mhd_file: Path,) -> SimpleITK.Image:
    headers = parse_mh_header(mhd_file)
    is_mha = headers["ElementDataFile"].strip() == "LOCAL"
    data_file_path = resolve_mh_data_file_path(headers, is_mha, mhd_file)

    shape = extract_header_listing("DimSize", headers=headers, dtype=int)

    dtype, num_components = determine_mh_components_and_dtype(headers)

    sitk_image = create_sitk_img_from_mh_data(
        data_file_path, dtype, headers, is_mha, num_components, shape
    )

    sitk_image.SetDirection(
        extract_header_listing("TransformMatrix", headers=headers)
    )
    sitk_image.SetSpacing(
        extract_header_listing("ElementSpacing", headers=headers)
    )
    sitk_image.SetOrigin(extract_header_listing("Offset", headers=headers))

    return sitk_image


def determine_mh_components_and_dtype(
    headers: Dict[str, str]
) -> Tuple[int, int]:
    num_components = 1
    if "ElementNumberOfChannels" in headers:
        num_components = int(headers["ElementNumberOfChannels"])
        if "_ARRAY" not in headers["ElementType"] and num_components > 1:
            headers["ElementType"] = headers["ElementType"] + "_ARRAY"
    dtype = METAIO_IMAGE_TYPES[headers["ElementType"]]
    if dtype is None:
        error_msg = (
            f"MetaIO datatype: {headers['ElementType']} is not supported"
        )
        raise NotImplementedError(error_msg)
    return dtype, num_components


def resolve_mh_data_file_path(
    headers: Dict[str, str], is_mha: bool, mhd_file: Path
) -> Path:
    if is_mha:
        data_file_path = mhd_file
    else:
        data_file_path = (
            mhd_file.resolve().parent / Path(headers["ElementDataFile"]).name
        )
    if not data_file_path.exists():
        raise OSError("cannot find data file")
    return data_file_path


def create_sitk_img_from_mh_data(
    data_file_path: Path,
    dtype: int,
    headers: Mapping[str, Union[str, None]],
    is_mha: bool,
    num_components: int,
    shape,
) -> SimpleITK.Image:
    is_compressed = headers["CompressedData"] == "True"
    with open(str(data_file_path), "rb") as f:
        if is_mha:
            line = b""
            while "ElementDataFile = LOCAL" not in str(line):
                line = f.readline()
        if not is_compressed:
            s = f.read()
        else:
            s = zlib.decompress(f.read())
    sitk_image = SimpleITK.Image(shape, dtype, num_components)
    _SimpleITK._SetImageFromArray(s, sitk_image)
    return sitk_image


def validate_and_clean_additional_mh_headers(
    headers: Dict[str, str]
) -> Dict[str, str]:
    cleaned_headers = {}
    for key, value in headers.items():
        if key in EXPECTED_HEADERS:
            cleaned_headers[key] = value
        else:
            if key in ADDITIONAL_HEADERS:
                match_pattern = ADDITIONAL_HEADERS[key]
                if not re.match(match_pattern, value):
                    raise ValueError(
                        f"Invalid data type found for "
                        f"additional header key: {key}"
                    )
                cleaned_headers[key] = value
        if key in HEADERS_MATCHING_NUM_TIMEPOINTS:
            validate_list_data_matches_num_timepoints(
                headers=headers, key=key, value=value
            )

    return cleaned_headers


def validate_list_data_matches_num_timepoints(
    headers: Dict[str, str], key: str, value: str
):
    num_timepoints = len(value.split(" "))
    expected_timepoints = (
        int(headers["DimSize"].split(" ")[3])
        if int(headers["NDims"]) >= 4
        else 1
    )
    if num_timepoints != expected_timepoints:
        raise ValueError(
            f"Found {num_timepoints} values for {key}, "
            f"but expected {expected_timepoints} (1/timepoint)"
        )


def add_additional_mh_headers_to_sitk_image(
    sitk_image: SimpleITK.Image, headers: Dict[str, str]
):
    cleaned_headers = validate_and_clean_additional_mh_headers(headers)
    for header in ADDITIONAL_HEADERS:
        if header in cleaned_headers:
            value = cleaned_headers[header]
            if isinstance(value, (list, tuple)):
                value = " ".join([str(v) for v in value])
            else:
                value = str(value)
            sitk_image.SetMetaData(header, value)


def load_sitk_image(mhd_file: Path) -> SimpleITK.Image:
    headers = parse_mh_header(mhd_file)
    headers = validate_and_clean_additional_mh_headers(headers=headers)
    ndims = int(headers["NDims"])
    if ndims < 4:
        sitk_image = SimpleITK.ReadImage(str(mhd_file))
        for key in sitk_image.GetMetaDataKeys():
            if key not in ADDITIONAL_HEADERS:
                sitk_image.EraseMetaData(key)
    elif ndims <= 4:
        sitk_image = load_sitk_image_with_nd_support(mhd_file=mhd_file)
    else:
        error_msg = (
            "SimpleITK images with more than 4 dimensions are not supported"
        )
        raise NotImplementedError(error_msg)
    add_additional_mh_headers_to_sitk_image(
        sitk_image=sitk_image, headers=headers
    )
    return sitk_image



from collections import defaultdict
from pathlib import Path
from typing import DefaultDict, Iterator, List, Set

import SimpleITK
import numpy as np
from PIL import Image
from PIL.Image import DecompressionBombError

from panimg.exceptions import UnconsumedFilesException, ValidationError
from panimg.models import SimpleITKImage


def format_error(message: str) -> str:
    return f"Fallback image builder: {message}"


def image_builder_fallback(*, files: Set[Path]) -> Iterator[SimpleITKImage]:
    """
    Constructs image objects by inspecting files in a directory.

    Parameters
    ----------
    path
        Path to a directory that contains all images that were uploaded during
        an upload session.

    Returns
    -------
    An `ImageBuilder` object consisting of:
     - a list of filenames for all files consumed by the image builder
     - a list of detected images
     - a list files associated with the detected images
     - path->error message map describing what is wrong with a given file
    """
    file_errors: DefaultDict[Path, List[str]] = defaultdict(list)

    for file in files:
        try:
            img = Image.open(file)

            if img.format.lower() not in ["jpeg", "png"]:
                raise ValidationError(
                    f"Unsupported image format: {img.format}"
                )

            img_array = np.array(img)
            is_vector = img.mode != "L"
            img = SimpleITK.GetImageFromArray(img_array, isVector=is_vector)

            yield SimpleITKImage(
                image=img,
                name=file.name,
                consumed_files={file},
                spacing_valid=False,
            )
        except (OSError, ValidationError, DecompressionBombError):
            file_errors[file].append(format_error("Not a valid image file"))

    if file_errors:
        raise UnconsumedFilesException(file_errors=file_errors)



from pathlib import Path
from typing import Dict, List


class ValidationError(Exception):
    pass


class UnconsumedFilesException(Exception):
    """
    Raised on completion of an image builder and there are unconsumed files.
    Contains a dictionary with a map of the errors encountered when loading
    the unconsumed file.
    """

    def __init__(self, *args, file_errors: Dict[Path, List[str]]):
        super().__init__(*args)
        self.file_errors = file_errors



from collections import defaultdict
from pathlib import Path
from typing import (
    DefaultDict,
    Iterable,
    Iterator,
    List,
    Set,
    Tuple,
    Union,
)

import SimpleITK
import numpy as np
from construct.core import (
    Float64l,
    Int8ul,
    PaddedString,
    StreamError,
    Struct,
)
from pydantic import BaseModel

from panimg.contrib.oct_converter.image_types import (
    FundusImageWithMetaData,
    OCTVolumeWithMetaData,
)
from panimg.contrib.oct_converter.readers import E2E, FDA, FDS
from panimg.exceptions import UnconsumedFilesException, ValidationError
from panimg.models import EyeChoice, SimpleITKImage


def format_error(message: str) -> str:
    return f"OCT image builder: {message}"


LATERALITY_TO_EYE_CHOICE = defaultdict(
    lambda: EyeChoice.UNKNOWN,
    {"L": EyeChoice.OCULUS_SINISTER, "R": EyeChoice.OCULUS_DEXTER},
)

OCT_CONVERTER_TYPE = Union[FDS, FDA, E2E]


class OctDimensions(BaseModel):
    extent_x_mm: float
    resolution_y_mm: float
    extent_z_mm: float


def _create_itk_images(
    *,
    file: Tuple[Path, str],
    oct_volumes: Iterable[OCTVolumeWithMetaData],
    fundus_images: Iterable[FundusImageWithMetaData],
    oct_slice_size: OctDimensions,
) -> Iterator[SimpleITKImage]:

    filepath = file[0]
    filetype = file[1]

    for volume in oct_volumes:
        eye_choice = LATERALITY_TO_EYE_CHOICE[volume.laterality]

        yield _create_itk_oct_volume(
            file=filepath,
            volume=volume.volume,
            oct_slice_size=oct_slice_size,
            eye_choice=eye_choice,
        )
    for image in fundus_images:
        eye_choice = LATERALITY_TO_EYE_CHOICE[image.laterality]

        if filetype != "e2e":
            img_array = image.image.astype(np.uint8)
            img_array = img_array[:, :, ::-1]
            is_vector = True
        else:
            img_array = image.image
            is_vector = False

        yield _create_itk_fundus_image(
            file=filepath,
            image=img_array,
            eye_choice=eye_choice,
            is_vector=is_vector,
        )


def _create_itk_oct_volume(
    *,
    file: Path,
    volume: List[np.ndarray],
    oct_slice_size: OctDimensions,
    eye_choice: EyeChoice,
) -> SimpleITKImage:
    img_array = np.array(volume)
    img = SimpleITK.GetImageFromArray(img_array, isVector=False)
    [st, _, sl] = img_array.shape
    img.SetSpacing(
        [
            oct_slice_size.extent_x_mm / sl,
            oct_slice_size.resolution_y_mm,
            oct_slice_size.extent_z_mm / st,
        ]
    )
    return SimpleITKImage(
        image=img,
        name=file.name,
        consumed_files={file},
        spacing_valid=True,
        eye_choice=eye_choice,
    )


def _create_itk_fundus_image(
    *, file: Path, image: np.ndarray, eye_choice: EyeChoice, is_vector: bool,
) -> SimpleITKImage:
    img = SimpleITK.GetImageFromArray(image, isVector=is_vector)

    return SimpleITKImage(
        image=img,
        name=file.stem + "_fundus" + file.suffix,
        consumed_files={file},
        spacing_valid=False,
        eye_choice=eye_choice,
    )


def _extract_slice_size(*, img: Union[FDS, FDA]) -> OctDimensions:
    slice_size_meta_data = Struct(
        "unknown" / PaddedString(12, "utf16"),
        "xmm" / Float64l,
        "zmm" / Float64l,
        "y" / Float64l,
        "unknown6" / Float64l,
        "unknown7" / Float64l,
        "unknown8" / Int8ul,
    )
    with open(img.filepath, "rb") as f:
        chunk_location, chunk_size = img.chunk_dict[b"@PARAM_SCAN_04"]
        f.seek(chunk_location)
        raw = f.read(chunk_size)
        spacing = slice_size_meta_data.parse(raw)
        return OctDimensions(
            extent_x_mm=spacing.xmm,
            resolution_y_mm=spacing.y / 1000.0,
            extent_z_mm=spacing.zmm,
        )


def _get_image(
    *, file: Path
) -> Tuple[
    Iterable[OCTVolumeWithMetaData],
    Iterable[FundusImageWithMetaData],
    OctDimensions,
    Tuple[Path, str],
]:
    with open(file, "rb") as f:
        header = f.read(7)

        if b"FDS" in header:
            fds_img = FDS(file)
            file_info = file, "fds"
            oct_slice_size = _extract_slice_size(img=fds_img)
            return (
                [fds_img.read_oct_volume()],
                [fds_img.read_fundus_image()],
                oct_slice_size,
                file_info,
            )
        elif b"FDA" in header:
            fda_img = FDA(file)
            file_info = file, "fda"
            oct_slice_size = _extract_slice_size(img=fda_img)
            return (
                [fda_img.read_oct_volume()],
                [fda_img.read_fundus_image()],
                oct_slice_size,
                file_info,
            )
        elif b"CMD" in header:
            e2e_img = E2E(file)
            file_info = file, "e2e"
            # We were unable to retrieve slice size information from the e2e files.
            # The following default values are taken from:
            # https://bitbucket.org/uocte/uocte/wiki/Heidelberg%20File%20Format
            oct_slice_size = OctDimensions(
                extent_x_mm=6, resolution_y_mm=0.0039, extent_z_mm=4.5
            )
            # Note that the return types from oct_converter are different
            # for e2e files
            return (
                e2e_img.read_oct_volume(),
                e2e_img.read_fundus_image(),
                oct_slice_size,
                file_info,
            )
        else:
            raise ValueError


def image_builder_oct(*, files: Set[Path]) -> Iterator[SimpleITKImage]:
    """
    Constructs OCT image objects by inspecting files in a directory by using
    OCT-converter PyPI library.

    Parameters
    ----------
    path
        Path to a directory that contains all images that were uploaded during
        an upload session.

    Returns
    -------
    An `ImageBuilder` object consisting of:
     - a list of filenames for all files consumed by the image builder
     - a list of detected images
     - a list files associated with the detected images
     - path->error message map describing what is wrong with a given file
    """
    file_errors: DefaultDict[Path, List[str]] = defaultdict(list)

    for file in files:
        try:
            oct_volumes, fundus_images, oct_slice_size, file_out = _get_image(
                file=file
            )

            yield from _create_itk_images(
                file=file_out,
                oct_volumes=oct_volumes,
                # Skip fundus_images until we decide how to handle these
                fundus_images=[],  # fundus_images,
                oct_slice_size=oct_slice_size,
            )

        except (OSError, ValidationError, StreamError, ValueError, IndexError):
            file_errors[file].append(
                format_error(
                    "Not a valid OCT file (supported formats: .fds,.fda,.e2e)"
                )
            )

    if file_errors:
        raise UnconsumedFilesException(file_errors=file_errors)



from collections import defaultdict
from pathlib import Path
from typing import DefaultDict, Iterator, List, Set

import SimpleITK

from panimg.exceptions import UnconsumedFilesException
from panimg.models import SimpleITKImage


def format_error(message: str) -> str:
    return f"NifTI image builder: {message}"


def image_builder_nifti(*, files: Set[Path],) -> Iterator[SimpleITKImage]:
    """
    Constructs image objects from files in NifTI format (nii/nii.gz)

    Parameters
    ----------
    files
        Path to images that were uploaded during an upload session.

    Returns
    -------
    An `ImageBuilder` object consisting of:
     - a list of filenames for all files consumed by the image builder
     - a list of detected images
     - a list files associated with the detected images
     - path->error message map describing what is wrong with a given file
    """
    file_errors: DefaultDict[Path, List[str]] = defaultdict(list)

    for file in files:
        if not (file.name.endswith(".nii") or file.name.endswith(".nii.gz")):
            file_errors[file].append(format_error("Not a NifTI image file"))
            continue

        try:
            reader = SimpleITK.ImageFileReader()
            reader.SetImageIO("NiftiImageIO")
            reader.SetFileName(str(file.absolute()))
            img: SimpleITK.Image = reader.Execute()
        except RuntimeError:
            file_errors[file].append(
                format_error("Not a valid NifTI image file")
            )
            continue

        yield SimpleITKImage(
            image=img,
            name=file.name,
            consumed_files={file},
            spacing_valid=True,
        )

    if file_errors:
        raise UnconsumedFilesException(file_errors=file_errors)



import logging
import shutil
from enum import Enum
from pathlib import Path
from typing import Dict, List, Optional, Set, Tuple
from uuid import UUID, uuid4

from SimpleITK import Image, WriteImage
from pydantic import BaseModel, validator
from pydantic.dataclasses import dataclass

logger = logging.getLogger(__name__)


class ColorSpace(str, Enum):
    GRAY = "GRAY"
    RGB = "RGB"
    RGBA = "RGBA"
    YCBCR = "YCBCR"


ITK_COLOR_SPACE_MAP = {
    1: ColorSpace.GRAY,
    3: ColorSpace.RGB,
    4: ColorSpace.RGBA,
}


class ImageType(str, Enum):
    MHD = "MHD"
    TIFF = "TIFF"
    DZI = "DZI"


class EyeChoice(str, Enum):
    OCULUS_DEXTER = "OD"
    OCULUS_SINISTER = "OS"
    UNKNOWN = "U"
    NOT_APPLICABLE = "NA"


@dataclass(frozen=True)
class PanImg:
    pk: UUID
    name: str
    width: int
    height: int
    depth: Optional[int]
    voxel_width_mm: Optional[float]
    voxel_height_mm: Optional[float]
    voxel_depth_mm: Optional[float]
    timepoints: Optional[int]
    resolution_levels: Optional[int]
    window_center: Optional[float]
    window_width: Optional[float]
    color_space: ColorSpace
    eye_choice: EyeChoice


@dataclass(frozen=True)
class PanImgFile:
    image_id: UUID
    image_type: ImageType
    file: Path


@dataclass(frozen=True)
class PanImgFolder:
    image_id: UUID
    folder: Path


@dataclass
class PanImgResult:
    new_images: Set[PanImg]
    new_image_files: Set[PanImgFile]
    new_folders: Set[PanImgFolder]
    consumed_files: Set[Path]
    file_errors: Dict[Path, List[str]]


@dataclass
class PostProcessorResult:
    new_image_files: Set[PanImgFile]
    new_folders: Set[PanImgFolder]


class SimpleITKImage(BaseModel):
    image: Image

    name: str
    consumed_files: Set[Path]

    spacing_valid: bool
    eye_choice: EyeChoice = EyeChoice.NOT_APPLICABLE

    class Config:
        arbitrary_types_allowed = True
        allow_mutation = False

    @property
    def width(self) -> int:
        return int(self.image.GetWidth())

    @property
    def height(self) -> int:
        return int(self.image.GetHeight())

    @property
    def depth(self) -> Optional[int]:
        try:
            depth = int(self.image.GetDepth())
        except (RuntimeError, ValueError):
            return None

        return depth or None

    @property
    def window_center(self) -> Optional[float]:
        try:
            return float(self.image.GetMetaData("WindowCenter"))
        except (RuntimeError, ValueError):
            return None

    @property
    def window_width(self) -> Optional[float]:
        try:
            return float(self.image.GetMetaData("WindowWidth"))
        except (RuntimeError, ValueError):
            return None

    @property
    def timepoints(self) -> Optional[int]:
        if self.image.GetDimension() == 4:
            return int(self.image.GetSize()[-1])
        else:
            return None

    @validator("image")
    def check_color_space(cls, image: Image):  # noqa: B902, N805
        cs = image.GetNumberOfComponentsPerPixel()
        if cs not in ITK_COLOR_SPACE_MAP:
            raise ValueError(f"Unknown color space for MetaIO image: {cs}")
        return image

    @property
    def color_space(self) -> ColorSpace:
        return ITK_COLOR_SPACE_MAP[self.image.GetNumberOfComponentsPerPixel()]

    @property
    def voxel_width_mm(self) -> Optional[float]:
        if self.spacing_valid:
            return float(self.image.GetSpacing()[0])
        else:
            return None

    @property
    def voxel_height_mm(self) -> Optional[float]:
        if self.spacing_valid:
            return float(self.image.GetSpacing()[1])
        else:
            return None

    @property
    def voxel_depth_mm(self) -> Optional[float]:
        if self.spacing_valid:
            try:
                return float(self.image.GetSpacing()[2])
            except IndexError:
                return None
        else:
            return None

    def save(self, output_directory: Path) -> Tuple[PanImg, Set[PanImgFile]]:
        pk = uuid4()

        work_dir = Path(output_directory) / self.name
        work_dir.mkdir()

        new_image = PanImg(
            pk=pk,
            name=self.name,
            width=self.width,
            height=self.height,
            depth=self.depth,
            window_center=self.window_center,
            window_width=self.window_width,
            timepoints=self.timepoints,
            resolution_levels=None,
            color_space=self.color_space,
            voxel_width_mm=self.voxel_width_mm,
            voxel_height_mm=self.voxel_height_mm,
            voxel_depth_mm=self.voxel_depth_mm,
            eye_choice=self.eye_choice,
        )

        WriteImage(
            image=self.image,
            fileName=str(work_dir.absolute() / f"{pk}.mha"),
            useCompression=True,
        )

        new_files = set()
        for file in work_dir.iterdir():
            new_file = PanImgFile(
                image_id=pk, image_type=ImageType.MHD, file=file,
            )
            new_files.add(new_file)

        return new_image, new_files


class TIFFImage(BaseModel):
    file: Path

    name: str
    consumed_files: Set[Path]

    width: int
    height: int
    voxel_width_mm: float
    voxel_height_mm: float
    resolution_levels: int
    color_space: ColorSpace
    eye_choice: EyeChoice = EyeChoice.NOT_APPLICABLE

    class Config:
        allow_mutation = False

    def save(self, output_directory: Path) -> Tuple[PanImg, Set[PanImgFile]]:
        pk = uuid4()

        output_file = output_directory / self.name / f"{pk}{self.file.suffix}"
        output_file.parent.mkdir()

        new_image = PanImg(
            pk=pk,
            name=self.name,
            width=self.width,
            height=self.height,
            depth=1,
            resolution_levels=self.resolution_levels,
            color_space=self.color_space,
            voxel_width_mm=self.voxel_width_mm,
            voxel_height_mm=self.voxel_height_mm,
            voxel_depth_mm=None,
            timepoints=None,
            window_center=None,
            window_width=None,
            eye_choice=self.eye_choice,
        )

        shutil.copy(src=self.file, dst=output_file)

        new_files = {
            PanImgFile(
                image_id=pk,
                image_type=ImageType.TIFF,
                file=output_file.absolute(),
            )
        }

        return new_image, new_files
